#include <p18cxxx.h>
#include <p18f452.h>
#include <xlcd.h>
#include <stdlib.h>
#include <delays.h>
#include <timers.h>
#include <capture.h>
#include "..\mylcd\mylcd.h"
#include "..\scrabblelcd\scrabblelcd.h"
#include "..\keypad\keypad.h"
#include "capturetest.h"

#define TRUE 1
#define FALSE 0

//Number of seconds passed between timer interrupts
#define SECONDS_PER_TIMECOUNT 	.25 //This is for the 1MHz clock.
									//For 30MHz, set this to (4.0/30)
									//and change the timer interrupt prescaler

//data capture flags
volatile char ccpflags = 0;

//Variables used in interrupts must be declared volatile
volatile unsigned int timecount = 0;	//timer interrupt count
volatile unsigned int buzzcount = 0;	//interrupt count for buzzer timing
volatile unsigned int button_press_val;	//holds vaue of button press
volatile unsigned int pause = FALSE;	//flag to determine pause status
volatile unsigned int flashticks = 0;	//total flash time in number of interrupts
volatile unsigned int flashon = FALSE;	//flag set if LED should be flashing

unsigned int flashval = 0;		//value of the flash bit (LED on or off)
unsigned int currentflash = 0;	//total time of current LED 
unsigned int haslimit = TRUE;	//flag to determine time limit status for current player
unsigned int timelimit = 60; 	//time limit of current player in seconds
unsigned int negscore = FALSE;	//flag to set pos/neg status of score
unsigned int timeleft;			//amount of time left for current player
unsigned int outoftime = FALSE;	//flag set if current player if out of time
unsigned int buzzed = FALSE;	//flag set if the buzzer has already gone off

int charnum=0;				 	//keeps track of position in entering data
int player=0;					//keeps track of the current player
char score_entry_string[10]; 	//stores entered score as string
int numplayers = 4;				//stores the number of players playing
player_info_t player_info[4];	//structure to hold player info

/*
* For PIC18cxxx devices the high interrupt vector is found at
* 00000008h. The following code will branch to the
* high_interrupt_service_routine function to handle
* interrupts that occur at the high vector.
*/
#pragma code high_vector=0x08
void interrupt_at_high_vector(void)
{
	//Poll each interrupt bit and branch to the correct routine
	//Have to use asm goto so that the function returns correctly
	if (INTCONbits.TMR0IF == 1) {
		_asm GOTO timer_isr _endasm
	}
	if (PIR1bits.CCP1IF == 1) {
		_asm GOTO capture1_isr _endasm
	}
	if (PIR2bits.CCP2IF == 1) {
		_asm GOTO capture2_isr _endasm
	}
}
#pragma code 

#pragma interrupt timer_isr save=PROD //Define timer_isr as an isr
void timer_isr (void)
{
	INTCONbits.TMR0IF = 0;//Turn the interrupt flag off
	if (pause == FALSE) {
		timecount++;	//Increment the time if not paused
	}
    if (flashon == TRUE) {
	    flashticks++;	//If the flasher is on, increment the flash time
	    currentflash++;	//
	}
	buzzcount++; //Increment the buzzer time
}

#pragma interrupt capture1_isr //define capture1_isr as an isr
void capture1_isr (void)
{
	//This interrupt is called when a keypad button is pressed
	PIR1bits.CCP1IF = 0;	//Turn the interrupt flag off
	registerPadPress();		//Store the current key press
	SET_CCP1;				//Set the flag indicating a key press
}

#pragma interrupt capture2_isr //define capture2_isr as an isr
void capture2_isr (void)
{
	//This interrupt is called when a large button is pressed
	PIR2bits.CCP2IF = 0;	//Turn the interrupt flag off
	SET_CCP2;				//Set the flag indicating a button press
	button_press_val = (PORTD & (0xf0)) >> 4;	//Calculate the index of the button
}

void main (void)
{
	char strtime[20];	//Character buffers to hold text data
	char buff[10];		//
	
	int charentered = FALSE;	//Has a character been entered (and not displayed)?
	char lastchar;				//Last character entered
	unsigned int redraw = FALSE;//Should a redraw be performed
	int key;					//Hold a key press value
	char time_entry_string[10];	//stores entered time as string
	int i;	//Commonly used in for loops
	
	int state = ENTER_PLAYERS_STATE;

	//initialize
	init();
	scrabbleLCDinit();
	setPlayerTurn(0);
	charnum = 0;

	TRISA = 0x00; 	//Set up port A as an output port
	PORTA = 0x00;	//Erase port A

	OpenTimer0 (TIMER_INT_ON & T0_SOURCE_INT & T0_16BIT & T0_PS_1_1); 	//set TO_PS_1_16 for
																		//for 30MHz clock 
	OpenCapture1 (CAPTURE_INT_ON & C1_EVERY_RISE_EDGE);	//Set up the capture ports
	OpenCapture2 (CAPTURE_INT_ON & C2_EVERY_RISE_EDGE);	//
	RCONbits.IPEN = 1;		//Turn on interrupt priority levels
	INTCONbits.GIEH = 1;	//Turn on high priority interrupts 
	
	//////////////////////////////////////////////////////////////
	//The main FSM loop.
	// We originally implemented this as a big switch statment but
	// some compiler optimization was causing strange errors (like
	// the code from the wrong state would randomly be run) so we
	// implemented as an if-elseif instead.
	while (1)
	{
		//{{{ ENTER_PLAYERS_STATE			
		//Get the number of players from the user
        if (state == ENTER_PLAYERS_STATE) {
			//Check for a captured keypress
			if (CCP1) {
				setPos(3, 35);
				key = getPadNum();
				//Make sure it's a valid number of players
				if (key < 5 && key > 1) {
                    numplayers = key;
					//Draw the screen for the next state
                    setPos(2, 0);
                    putrsXLCD("   Enter the initials of each player.   ");
                   	putrsXLCD("      Press TIME to enter times. ");
                    dispScores(player_info, numplayers);
                    placeCursor(0,0);
                    state = ENTER_NAMES_STATE;	//Switch states
				}
				CLR_CCP1; //Clear the capture flag
			}
        }
		//}}}
		//{{{ ENTER_NAMES_STATE
		// Get the initials of each player
		else if (state == ENTER_NAMES_STATE) {
			//keypress captured
			if (CCP1) {
				//If enter is pressed store the current character and advance
				if (getPadNum()==ENTER_KEY) {
					if (charentered == TRUE) {
						player_info[player].name[charnum] = lastchar;
						charentered = FALSE;
					}
					charnum++;
					
					//Advance to the next player if past 3 characters
					if (charnum==3) {
						player++;
						if (player>numplayers-1) player=0;
						charnum=0;
						setPlayerTurn(player);
					}
					placeCursor(player, charnum);
				}
				
				//If backspace is pressed, erase the current character and go back.
				else if (getPadNum()==BACKSPACE_KEY) {
					placeChar(player, charnum, ' ');
					charnum--;
					if (charnum<0) {
						player--;
						charnum=2;
						setPlayerTurn(player);
					}
					if (player<0) player = numplayers-1;
					placeCursor(player, charnum);
				}

				//If not enter or backspace then this is a character
				//Write the char to the screen and buffer it
				else {
					charentered = TRUE;
					lastchar = nextChar();
					placeChar(player, charnum, lastchar);
				}
				CLR_CCP1; //Clear the keypress flag
			}

			//Button press captured
			if (CCP2) {
				CLR_CCP2; //Clear the button press flag
				switch (button_press_val) {
					case TIME_BUTTON:	//Done entering names, advance to the time state
						charnum = 0;
						clearLine(2);
						scrabbleLCDinitSetTimeMode();
						state = SET_TIMES_STATE;
						break;
					
					case TIME_BYPASS_BUTTON:	//Done entering names, advance straight to
					case START_BUTTON:			//gameplay. No timelimits will be entered
						initRunMode(0);
						scrabbleLCDinitRunMode();
						state = RUN_STATE;
						break;
						
					case PLAYER1_BUTTON:		//Move the cursor to player 1
						if (charentered == TRUE) {
							player_info[player].name[charnum] = lastchar;
							charentered = FALSE;
						}
						charnum = 0;
						player = 0;
						placeCursor(player, charnum);
						break;
						
					case PLAYER2_BUTTON:		//Move the cursor to player 2
						if (charentered == TRUE) {
							player_info[player].name[charnum] = lastchar;
							charentered = FALSE;
						}
						player = 1;
						charnum = 0;
						placeCursor(player, charnum);
						break;
						
					case PLAYER3_BUTTON:		//Move the cursor to player 3
						if (charentered == TRUE) {
							player_info[player].name[charnum] = lastchar;
							charentered = FALSE;
						}
						charnum = 0;
						player = 2;
						placeCursor(player, charnum);
						break;
						
					case PLAYER4_BUTTON:		//Move the cursor to player 4
						if (charentered == TRUE) {
							player_info[player].name[charnum] = lastchar;
							charentered = FALSE;
						}
						charnum = 0;
						player = 3;
						placeCursor(player, charnum);
						break;
				}
			}
		}
		//}}}
		//{{{ SET_TIMES_STATE
		// Set the time limits for each player
		else if (state == SET_TIMES_STATE) {
			if (CCP1) {
				setPos(3, TIME_ALL_MODE_TIME_POS + charnum);
				key = getPadNum();

				if (key==ENTER_KEY){
					//No effect
					//Must hit a player button to assign a time
				}
				//Erase the current character and go back
				else if (key==BACKSPACE_KEY) {
					putchar(' ');
					time_entry_string[charnum] = ' ';
					charnum--;

					if (charnum<0) { charnum = 0; }
					setPos(3, TIME_ALL_MODE_TIME_POS + charnum);
					putchar(' ');
				}
				//If not enter or backspace, must be a number
				//Display the number and buffer it
				else {
					if (charnum < 5) {
						putchar(key + '0');
						time_entry_string[charnum] = (key + '0');
						charnum++;
					}
				}
				CLR_CCP1; //Clear the keypress flag
			}

			//Button press captured
			if (CCP2) {
				CLR_CCP2; //Clear button press flag
				switch (button_press_val) {
					time_entry_string[charnum] = '\0'; //Null terminator for string

					//Assign the time limit to all players
					case ALL_BUTTON:
						for (i = 0; i < numplayers; i++){
							player_info[i].timelimit = atoi(time_entry_string);
						}
						redraw = TRUE;
						break;

					//Assign the time limit to player 1
					case PLAYER1_BUTTON:
						player_info[0].timelimit = atoi(time_entry_string);
						redraw = TRUE;
						break;

					//Assign the time limit to player 2
					case PLAYER2_BUTTON:
						player_info[1].timelimit = atoi(time_entry_string);
						redraw = TRUE;
						break;

					//Assign the time limit to player 3
					case PLAYER3_BUTTON:
                        if (numplayers < 3) break;
						player_info[2].timelimit = atoi(time_entry_string);
						redraw = TRUE;
						break;

					//Assign the time limit to player 4
					case PLAYER4_BUTTON:
                        if (numplayers < 4) break;
						player_info[3].timelimit = atoi(time_entry_string);
						redraw = TRUE;
						break;

					//Changed your mind, don't want to enter times. Bypass timing altogether
					//and advance right into gameplay
					case TIME_BYPASS_BUTTON:
						for (i=0; i < numplayers; i++) {
							player_info[i].timelimit = 0;
						}
					//Don't break, still need to start gameplay
					case START_BUTTON:
						initRunMode(0);
						scrabbleLCDinitRunMode();
						state = RUN_STATE;
				} //end switch(button_press_val)

				//display new time limits on line 2
				if (redraw) {
					charnum = 0;
					setPos(3, TIME_ALL_MODE_TIME_POS);
					putrsXLCD("     ");
					redraw = FALSE;
					clearLine(2);
					for (i = 0; i < numplayers; i++) {
                        if (player_info[i].timelimit == 0)
                            continue; //No time for some players
						setPos(2, i*10 + 3);
						putrsXLCD("     ");
						itoa(player_info[i].timelimit, buff);
						setPos(2, i*10 + 3);
						putsXLCD(buff);
					}
					setPos(3, TIME_ALL_MODE_TIME_POS + charnum);
				}
			}
		}
//}}}
		//{{{ RUN_STATE
		// Perform actual gameplay
		else if (state == RUN_STATE) {
			redraw = TRUE;	//Need to redraw the time

			//Check if the flashing led is on.
            if (flashon) {
				//Alternate between on for FLASH_RATE_ON and off for FLASH_RATE_OFF
                if (flashval == 1 && currentflash * SECONDS_PER_TIMECOUNT > FLASH_RATE_ON) {
	                currentflash = 0;
	                flashval = 0;
	                PORTA = 0x00;
	            }
	            else if (flashval == 0 && currentflash * SECONDS_PER_TIMECOUNT > FLASH_RATE_OFF) {
		            currentflash = 0;
		            flashval = 1;
		            PORTA = 0b00000001;
		        }
            }
            else {
				//Just in case
	            PORTA=0x00;
            }

			//If this player has a time limit and the timer is not paused then
			//check all time constraints and update the display
			if ( haslimit && !pause && ((int)(timelimit - timecount*SECONDS_PER_TIMECOUNT)) < timeleft ) {
				timeleft = timelimit - timecount*SECONDS_PER_TIMECOUNT;

				//Turn the flasher on when there are FLASH_START seconds left
                if (timeleft <= FLASH_START && !flashon) {
                    flashon = TRUE;
                    currentflash = 0;
                    flashval = 1;
                    PORTA = 1;
                }

				//Out of time, perform necessary operations
				if (timeleft <= 0) {
					do_outoftime();
					redraw = FALSE;
					//Redraw once more now
					dispTimeSecs(player, timeleft);
					dispTimeBarRatio((float)timeleft/(float)timelimit);
					timeleft = 0;
					//If we haven't buzzed already, start the buzzer
					if (!buzzed) {
						state = BUZZ_STATE;
						buzzed = TRUE;
						buzzcount = 0;
					}
					//We need to wait for score entry before we advance to the next player
				}

				//Redraw the time status
				if (redraw == TRUE) {
					redraw = FALSE;
					dispTimeSecs(player, timeleft);
					dispTimeBarRatio((float)timeleft/(float)timelimit);
				}
			}

			//Key press capture
			if (CCP1) {
				setPos(2, ENTER_SCORE_VALUE_POS + 1 + charnum);
				key = getPadNum();

				//If enter is pressed, apply the score to the current player 
				if (key==ENTER_KEY && charnum > 0){
					score_entry_string[charnum] = '\0';
					player_info[player].last_score = atoi(score_entry_string);
                    if (negscore) { //Adjust for negative score entry
                        player_info[player].last_score = -player_info[player].last_score;
                    }
				    player_info[player].score += player_info[player].last_score;
                    negscore = FALSE;
					dispScores(player_info, numplayers); //Update the display
					initRunMode((player+1)%numplayers);
					scrabbleLCDinitRunMode();
				}

				//If backspace is pressed erase the last digit. If there are no more digits
				//then toggle the negative operator
				else if (key==BACKSPACE_KEY) {
					charnum--;
					score_entry_string[charnum] = ' ';
					setPos(2, ENTER_SCORE_VALUE_POS + 1 + charnum);
					putchar(' ');
					if (charnum<0) { 	//Toggle negative
                        setPos(2, ENTER_SCORE_VALUE_POS + 1 + charnum);
                        if (negscore) {
                            negscore = FALSE;
                            putchar(' ');
                        }
                        else
                        {
                            negscore = TRUE;
                            putchar('-');
                        }
                        putchar(' ');
                        charnum = 0;
                    }
					setPos(2, ENTER_SCORE_VALUE_POS + 1 + charnum);
				}

				//Otherwise, if the key isn't enter it's a digit
				//Append the digit to the current number
				else if (key != ENTER_KEY) {
					if (charnum < 5) {
						putchar(key + '0');
						score_entry_string[charnum] = (key + '0');
						charnum++;
					}
				}
				
				CLR_CCP1; //Clear the keypress flag
			}

			//Button press event
			if (CCP2) {
				CLR_CCP2;//Clear the button press flag
				switch (button_press_val) {

					//The start button unpauses if paused
					case START_BUTTON:
						unpause();
						break;

					//The pause button pauses if not paused
					case PAUSE_BUTTON:
						dopause();
						break;

					//Apply the currently entered score to player 1.
					//If it is player 1's turn advance to the next player
					case PLAYER1_BUTTON:
                        if (charnum == 0) break;
						score_entry_string[charnum] = '\0';
						player_info[0].last_score = atoi(score_entry_string);
                        if (negscore) {
                            player_info[0].last_score = -player_info[0].last_score;
                        }
						player_info[0].score += player_info[0].last_score;
						redraw = TRUE;
                        negscore = FALSE;
                        charnum = 0;	//Start a new string
                        if (player==0) { 	//Advance to next player
    					    initRunMode((player+1)%numplayers);
        					scrabbleLCDinitRunMode();
                        }
						break;

					//Apply the currently entered score to player 2.
					//If it is player 2's turn advance to the next player
					case PLAYER2_BUTTON:
                        if (charnum == 0) break;
						score_entry_string[charnum] = '\0';
						player_info[1].last_score = atoi(score_entry_string);
                        if (negscore) {
                            player_info[1].last_score = -player_info[1].last_score;
                        }
						player_info[1].score += player_info[1].last_score;
						redraw = TRUE;
                        negscore = FALSE;
                        charnum = 0;
                        if (player==1) { 
    					    initRunMode((player+1)%numplayers);
        					scrabbleLCDinitRunMode();
                        }
						break;

					//Apply the currently entered score to player 3.
					//If it is player 3's turn advance to the next player
					case PLAYER3_BUTTON:
                        if (charnum == 0) break;
                        if (numplayers < 3) break; //Check that there are indeed at least 3 players
						score_entry_string[charnum] = '\0';
						player_info[2].last_score = atoi(score_entry_string);
                        if (negscore) {
                            player_info[2].last_score = -player_info[2].last_score;
                        }
						player_info[2].score += player_info[2].last_score;
						redraw = TRUE;
                        negscore = FALSE;
                        charnum = 0;
                        if (player==2) { 
    					    initRunMode((player+1)%numplayers);
        					scrabbleLCDinitRunMode();
                        }
						break;

					//Apply the currently entered score to player 14
					//If it is player 4's turn advance to the next player
					case PLAYER4_BUTTON:
                        if (charnum == 0) break;
                        if (numplayers < 4) break;	//Check that there are 4 players
						score_entry_string[charnum] = '\0';
						player_info[3].last_score = atoi(score_entry_string);
                        if (negscore) {
                            player_info[3].last_score = -player_info[2].last_score;
                        }
						player_info[3].score += player_info[3].last_score;
						redraw = TRUE;
                        negscore = FALSE;
                        charnum = 0;
                        if (player==3) { 
    					    initRunMode((player+1)%numplayers);
        					scrabbleLCDinitRunMode();
                        }
						break;

					//Don't want to use the timer anymore
					//Eliminate the timer for the current player as well as all other players
                    case TIME_BYPASS_BUTTON:
                    	for (i=0; i < numplayers; i++) {
							player_info[i].timelimit = 0;
						}
                        nolimit();
                        haslimit = FALSE;
                        break;

					//Restart the timer for the current player
                    case RESET_TIMER_BUTTON:
                    	initRunMode(player);
                    	break;
				}

				//Redraw the score
				if (redraw) {
					redraw = FALSE;
					dispScores(player_info, numplayers);
					clearLine(2);
					if (pause) {
						setPos(2,0);
						putrsXLCD("PAUSED");
					}
					setPos(2, ENTER_SCORE_VALUE_POS - 13);
					if (negscore)
						putrsXLCD("Enter Score:-");
					else
						putrsXLCD("Enter Score:");											
					setPos(2, ENTER_SCORE_VALUE_POS + 1);
					charnum = 0;
				}
			}
		}
		//}}}
		//{{{ BUZZ_STATE
		// Sound the buzzer for some time then return to the run state
		else if (state == BUZZ_STATE) {
			//Turn buzzer on
			PORTA |= 0b00000100;
			timeleft = BUZZ_LENGTH - buzzcount*SECONDS_PER_TIMECOUNT;
			if (timeleft <= 0) {
				//buzz time up, turn buzzer off
				PORTA &= 0b11111011;
				timeleft = 0;
				state = RUN_STATE;
			}
		}
		//}}}
	} // end while
}

//Initialize the player info structs
void init(void) {
	int i;
	for (i=0; i<numplayers; i++) {
		player_info[i].name[0] = '\0';
		player_info[i].timelimit = 0;
		player_info[i].last_score = 0;
		player_info[i].score = 0;
		player_info[i].bingos=0;
	}
}

//Initialze the gameplay state for the current player
void initRunMode(int playernum) {
	charnum = 0;
	player = playernum;
    flashticks = 0;
    flashon = FALSE;
	setPlayerTurn(player);
	timecount=0;
	timelimit = player_info[playernum].timelimit;
	timeleft = player_info[playernum].timelimit;
    if (timelimit == 0)
        haslimit = FALSE;
    else
        haslimit = TRUE;
	buzzed = FALSE;
	un_outoftime();	
}

//Turn the pause flag on
void dopause(void) {
	pause = TRUE;
	setPos(2, 0);
	putrsXLCD("PAUSED");
}

//Turn the pause flag off
void unpause(void) {
	pause = FALSE;
	//setPos(2, TIME_START_COL + 7);
	setPos(2, 0);
	putrsXLCD("      ");
}

//Toggle the pause flag
void togglepause(void) {
	pause = 1-pause;
	setPos(2, 0);
	if(pause==TRUE) {
		putrsXLCD("PAUSED");
	} else {
		putrsXLCD("      ");
	}
}

//If out of time, write it to the lcd
void do_outoftime(void) {
	outoftime = TRUE;
	setPos(3, 20);
	putrsXLCD("OUT OF TIME");
}

//Clear the out of time text
void un_outoftime(void) {
	outoftime = FALSE;
	setPos(3, 20);
	putrsXLCD("           ");
}

//Set up the screen for a player with no time limit
void nolimit(void) {
    setPos(3, 19);
    putrsXLCD("NO TIME LIMIT");
}
